module HibanaSemantics where

    import HibanaTypes
    import HibanaFunctions

    evalStatements :: State -> Statements -> State
    -- Basic Statement Rules
    evalStatements env Null = env
    evalStatements (StError e) _ = (StError e)
    evalStatements env (Composition st1 st2) = let  
                            			env' = evalStatements env st1
                            			env'' = evalStatements env' st2
					       in env''
    evalStatements env (AssignVar x e) = let k = evalExpr env e in
                        if variable_type k env x then
			    upd_v env x k
                        else
			    StError Constraint_Error
    evalStatements env (AssignArr a le e) = 
			let lk = evalIndexes env le in
                        if type_array_ind env a lk then
                            let k = evalExpr env e in
                            if type_array_val env a k then
                                upd_a env a lk k
                            else StError Constraint_Error
                        else StError Constraint_Error
    -- Basic Control Structures
    evalStatements env (IfThenElse b s1 s2) = 
			let bk = evalBExpr env b
			in if isTError bk then let (TError err) = bk in StError err 
			    else if isTrue bk then evalStatements env s1 
				else evalStatements env s2
    evalStatements env (While as b s) = 
		    let bk = evalBExpr env b
                    in if isTError bk then let (TError err) = bk in StError err
			else if isTrue bk then 
				let env' = evalStatements env s
				    env'' = evalStatements env' (While as b s)
				in env''
			     else env
    -- Extended control structures based on rewriting
    evalStatements env (If b s) = let env' = evalStatements env (IfThenElse b s Null) in env'
    evalStatements env (IfElsIf b1 s1 b2 s2) = let env' = evalStatements env (IfThenElse b1 s1 (If b2 s2)) in env'
    evalStatements env (IfElsIfElse b1 s1 b2 s2 sf) = let env' = evalStatements env (IfThenElse b1 s1 (IfThenElse b2 s2 sf)) in env'
    evalStatements env (IfListElsIf b0 s0 [(bf, sf)]) = let env' = evalStatements env (IfElsIf b0 s0 bf sf) in env'
    evalStatements env (IfListElsIf b0 s0 ((b1,s1):xs)) = let env' = evalStatements env (IfThenElse b0 s0 (IfListElsIf b1 s1 xs)) in env'
    evalStatements env (IfListElsIfElse b0 s0 [(bn, sn)] sf) = let env' = evalStatements env (IfElsIfElse b0 s0 bn sn sf) in env'
    evalStatements env (IfListElsIfElse b0 s0 ((b1,s1):xs) sf) = let env' = evalStatements env (IfThenElse b0 s0 (IfListElsIfElse b1 s1 xs sf)) in env'
    evalStatements env (InfiniteLoop s) = let env' = evalStatements env (While Assert (BEAtom True) s) in env'
    evalStatements env (LoopExitWhen1 s b) = let
                             env' = evalStatements env s
                             env'' = evalStatements env' (While Assert (BENot b) s)
                        in if isError env' then env' else env''
    evalStatements env (LoopExitWhen2 as b s) = let env' = evalStatements env' (While as (BENot b) s) in env'
    -- For Loop is missing
    -- Rules for Declarations
    evalStatements env (Declaration_Var dv) = evalVarDeclarations env dv
    evalStatements env (Declaration_Function df) = evalFunctionDeclarations env df
    evalStatements env (Declaration_Subprogram dsp) = evalSubprogramDeclarations env dsp
    -- Subprogram Invocation
    evalStatements env (CallPP p) = let env'' = evalStatements env' s in
                    if isError env'' then env'' else
                        let env''' = restoreLocalVariablesP env env'' p
                        in env''' 
                where
                env' = getEnvP env p
                s = getBodyProcedure env p
    evalStatements env (CallPF f) = let 
                    env'' = evalStatements env' s
                    k = evalExpr env'' exp
                    env''' = restoreLocalVariablesF env env'' f
                    in upd_v env''' (idftoidv f) k
                where
                env' = getEnvF env f
                s = getBodyFunction env f
                exp = getReturnExpression env f
    -- mSPARK entry point
    evalStatements env (Execution_Block dt dv dsp s) = let 
                                env' = evalTypeDeclarations env dt 
                                env'' = evalVarDeclarations env' dv
                                env''' = evalSubprogramDeclarations env'' dsp
                                env'''' = evalStatements env''' s
                                in if isError env' then env' 
                                    else if isError env'' then env'' else
                                        if isError env''' then env''' else env''''

    -- Variable declaration rules
    evalVarDeclarations :: State -> D_v -> State
    evalVarDeclarations (StError e) _ = StError e
    evalVarDeclarations env End_D_v = env
    evalVarDeclarations env (Dv [x] t dv) = let env' = addVarAssign env x t Empty
						env'' = evalVarDeclarations env' dv
					    in env''
    evalVarDeclarations env (Dv (x:xs) t dv) = let env' = addVarAssign env x t Empty
						   env'' = evalVarDeclarations env' (Dv xs t dv)
					       in env''
    evalVarDeclarations env (DAv [x] t cae dv) = let k = evalCAExpr cae in
						   if isTError k then let (TError e) = k in StError e
						   -- Lacks Range Validation
						   else	let
							    env' = addVarAssign env x t k
							    env'' = evalVarDeclarations env' dv
							in env''
    evalVarDeclarations env (DAv (x:xs) t cae dv) = let k = evalCAExpr cae in
						   if isTError k then let (TError e) = k in StError e
						   -- Lacks Range Validation
						   else	let
							    env' = addVarAssign env x t k
							    env'' = evalVarDeclarations env' (DAv xs t cae dv)
							in env''
                

    -- Declaration of Functions    	
    evalFunctionDeclarations :: State -> D_f -> State
    evalFunctionDeclarations (StError e) _ = StError e
    evalFunctionDeclarations env End_D_f = env
    evalFunctionDeclarations env (Dpf f c t dv df1 s e df2) = let env' = addFunctionAndVar env f t
								  -- Need types to check initialization of variables
								  env0 = reduceState env 
								  env1 = evalVarDeclarations env0 dv
								  env2 = evalFunctionDeclarations env1 df1
								  env3 = upd_f f c t s e empty_vmap env2 env'
								  envf = evalFunctionDeclarations env3 df2
							      in envf
    evalFunctionDeclarations env (Df f par c t dv df1 s e df2) = let env' = addFunctionAndVar env f t
								  -- Need types to check initialization of variables
								     env0 = reduceState env 
								     env1 = evalVarDeclarations env0 dv
								     env2 = evalFunctionDeclarations env1 df1
								     env3 = upd_f f c t s e par env2 env'
								     envf = evalFunctionDeclarations env3 df2
								 in envf
        

    evalProcedureDeclarations :: State -> D_p -> State
    evalProcedureDeclarations (StError e) _ = StError e
    evalProcedureDeclarations env End_D_p = env
    evalProcedureDeclarations env (Dpp p c dv dsp1 s dsp2) = let env' = addProcedure p env
							         env'' = evalVarDeclarations env' dv
							         env''' = evalSubprogramDeclarations env'' dsp1
							         env4 = upd_p p c s [] env'''
							         env5 = evalSubprogramDeclarations env4 dsp2
							     in env5

    evalSubprogramDeclarations :: State -> D_sp -> State
    evalSubprogramDeclarations env (D_spf df) = evalFunctionDeclarations env df
    evalSubprogramDeclarations env (D_spp dp) = evalProcedureDeclarations env dp

    evalTypeDeclarations :: State -> D_t -> State
    evalTypeDeclarations env End_D_t = env
    evalTypeDeclarations env (Dt t r dt) = let env' = addRangeType env t r
					       env'' = evalTypeDeclarations env' dt
					   in env''
    evalTypeDeclarations env (Dst t bt r dt) = let env' = addRangeSubtype env t bt r
						   env'' = evalTypeDeclarations env' dt
					       in env''
