module HibanaTypes where

    type List a = [a]
    data NEList a = One a | Cons a (NEList a)
        deriving (Show, Eq)
    type ID = String

    data State = State (TMap, VMap, PMap, FMap) | StError Error

    type TMap = (List ID_t, RT)
    type VMap = (List ID_v, V)
    type PMap = (List ID_p, P)
    type FMap = (List ID_f, F)

    toStringTMap :: TMap -> String
    toStringTMap ([], _) = []
    toStringTMap (x:xs, RT f) = (show x)++": "++(show (f x))++(toStringTMap (xs, RT f))

    toStringVMap :: VMap -> String
    toStringVMap ([], _) = []
    toStringVMap (x:xs, V f) = (show x)++" = "++(show (f x))++"\n"++(toStringVMap (xs,V f))

    toStringFMap :: FMap -> String
    toStringFMap ([], _) = []
    toStringFMap (x:xs, F f) = (show x)++(auxPrint (f x))++"\n"++(toStringFMap (xs, F f))
	    where
	    auxPrint (Nothing) = " = Nothing"
	    auxPrint (Just (c, p, t, (vmap, fmap), s, e)) = " definition\nContract = "++(show c)++"\nParameters = "++(toStringVMap p)++"\nReturn type: "++(show t)++"\nLocal Variables\n"++(toStringVMap vmap)++"\nBody\n"++(show s)++"\nReturn exp = "++(show e)++"\nNested Functions\n"++(toStringFMap fmap)++"\nEnd of definition of "++(show x)

    instance Show State where
	show (StError e) = "Error State: "++(show e)++"\n"
	show (State (t, v, p, f)) = "types\n"++ (toStringTMap t) ++ "\nVariables\n"++(toStringVMap v)++"\nFunctions\n"++(toStringFMap f)++"\n"

    empty_state :: State
    empty_state = State (empty_tmap, empty_vmap, empty_pmap, empty_fmap)

    empty_tmap = ([], RT (\_ -> Nothing))
    empty_vmap = ([], V (\_ -> Nothing))
    empty_pmap = ([], P (\_ -> Nothing ))
    empty_fmap = ([], F (\_ -> Nothing ))

    data Error = Error |
		 Constraint_Error 
        deriving Show

    data V = V (ID_v -> Maybe (ID_t, T))
    data T = TError Error | 
	     Empty | 
	     TBool Bool | 
	     TChar Char | 
	     TInt Int | 
	     TReal Float | 
	     TString String | 
	     TArray (R, List (NEList T), (NEList T -> T))

    instance Show T where
	show (TError e) = "Error: "++(show e)
	show Empty = "Unitialized Value"
	show (TBool b) = show b
	show (TChar c) = show c
	show (TInt i) = show i
	show (TReal f) = show f
	show (TString s) = s
	show (TArray f) = "Array map"

    instance Eq T where
	(==) (TInt i1) (TInt i2) = i1==i2
	(==) (TReal f1) (TReal f2) = f1==f2
	(==) (TArray f1) (TArray f2) = True
    -- Equality of arrays needs to be implemented

    data ID_v = ID_v ID
	deriving Eq
    
    instance Show ID_v where
        show (ID_v v) = v

    data ID_t = ID_t ID
	deriving Eq

    instance Show ID_t where
	show (ID_t id) = id

    data P = P (ID_p -> Maybe (ContractSpec, Parameters_p, Local_p, Statements))
    type Parameters_p = (List (Mode, ID_v), V)
    data Mode = In | Out | In_Out
        deriving Show
    type Local_p = (VMap, PMap, FMap)
    data ID_p = ID_p ID
	deriving Eq

    instance Show ID_p where
        show (ID_p p) = p
    
    data F = F (ID_f -> Maybe (ContractSpec, Parameters_f, ID_t, Local_f, Statements, E))
    
    showFunctions :: FMap -> String
    showFunctions ([], F f) = " "
    showFunctions (x:xs, F f) = "function  f = " ++(show (f x))++"\n"
    
    type Parameters_f = VMap
    type Local_f = (VMap, FMap)
    data ID_f = ID_f ID
	deriving Eq

    instance Show ID_f where
        show (ID_f f) = f

    data RT = RT (ID_t -> Maybe (Either (NEList ID_t) (Maybe ID_t, R)))

    instance Show RT where
	show (RT f) = "RT Function"

    data R = RInt Int Int | RChar Char Char | REnum ID_t ID_t | RArray (NEList R) R
	deriving Eq

    instance Show R where
	show (RInt a b) = "range "++(show a)++" .. "++(show b)
	show (RChar a b) = "range "++(show a)++" .. "++(show b)
	show (REnum id1 id2) = "range "++(show id1)++" .. "++(show id2)
	show (RArray nelr t) = "array (" ++ (printNEList nelr) ++ ") of "++(show t)

    printNEList (One r) = show r
    printNEList (Cons h t) = (show h)++", "++(printNEList t)

    data E = E |
	     Boolean BE |
	     K CAE |
	     Sum E E |
	     Mul E E |
	     Div E E |
	     Sub E E |
	     Paren E |
	     Var ID_v [E] |
	     FCall ID_f [E] |
	     Atom T 

    instance Show E where
	show E = "E"
	show (Boolean be) = show be
	show (K cae) = show cae
	show (Sum e1 e2) = (show e1)++" + "++(show e2)
	show (Mul e1 e2) = (show e1)++" * "++(show e2)
	show (Div e1 e2) = (show e1)++" / "++(show e2)
	show (Sub e1 e2) = (show e1)++" - "++(show e2)
	show (Paren e) = "( "++(show e)++" )"
	show (Var v []) = show v
	show (Var v le) = (show v )++ "( "++ (printList le) ++ " )"
	show (FCall f le) = (show f) ++ "( " ++ (printList le) ++ " )"
	show (Atom t) = show t

    data BE = BE | 
	      BEAtom Bool | 
	      BENot BE |
	      BEParen BE |
	      And BE BE |
	      Or BE BE |
	      BGT E E |
	      GTE E E |
	      BLT E E |
	      LTE E E |
	      Eq E E

    instance Show BE where
	show BE = "BE"
	show (BEAtom b) = show b
	show (BENot b) = "not "++(show b)
	show (BEParen b) = "( "++(show b)++" )"
	show (And b1 b2) = (show b1)++" and "++(show b2)
	show (Or b1 b2) = (show b1)++" or "++(show b2)
	show (BGT e1 e2) = (show e1)++" > "++(show e2)
	show (GTE e1 e2) = (show e1)++" >= "++(show e2)
	show (BLT e1 e2) = (show e1)++" < "++(show e2)
	show (LTE e1 e2) = (show e1)++" <= "++(show e2)
	show (Eq e1 e2) = (show e1)++" = "++(show e2)

    data CAE = CAE |
	       KSum CAE CAE |
	       KMul CAE CAE |
	       KDiv CAE CAE |
	       KSub CAE CAE |
	       KParen CAE |
	       KAtom T 

    instance Show CAE where
	show CAE = "CAE"
	show (KSum k1 k2) = (show k1)++" + "++(show k2)
	show (KMul k1 k2) = (show k1)++" * "++(show k2)
	show (KDiv k1 k2) = (show k1)++" / "++(show k2)
	show (KSub k1 k2) = (show k1)++" - "++(show k2)
	show (KParen k) = "( "++(show k)++" )"
	show (KAtom t) = show t
	       	
    data AssertExpr = AssertExpr |
		      BExpr BE |
		      LImp BE BE |
		      LEq BE BE
	deriving Show

    data Assert = Assert |
	          Empty_Assert |
		  ForAll (List ID_v) AssertExpr |
		  Exists (List ID_v) AssertExpr
        deriving Show

    data ContractSpec = ContractFunction Assert Assert |
			ContractProcedure Assert Assert
	deriving Show

    data Global = Global (List ID_v)
	deriving Show

    instance Show V where
        show _ = "V"

    instance Show F where
        show _ = "F"

    data Statements = Null | 
              Composition Statements Statements |
              AssignVar ID_v E |
              AssignArr ID_v [E] E |
              IfThenElse BE Statements Statements |
              While Assert BE Statements |
              If BE Statements |
              IfElsIf BE Statements BE Statements |
              IfElsIfElse BE Statements BE Statements Statements |
              IfListElsIf BE Statements [(BE, Statements)] |
              IfListElsIfElse BE Statements [(BE, Statements)] Statements |
              InfiniteLoop Statements |
              LoopExitWhen1 Statements BE |
              LoopExitWhen2 Assert BE Statements |
              Declaration_Var D_v |
              Declaration_Function D_f |
              Declaration_Subprogram D_sp |
              Declaration_Type D_t |
              CallPF ID_f |
	      CallF ID_f [E] |
              CallPP ID_p |
              Execution_Block D_t D_v D_sp Statements

    instance Show Statements where
        show s = (prettyPrintStatements s 0)++";\n"

    printTabs :: Int -> String
    printTabs 0 = []
    printTabs n = pt++(printTabs (n-1))
        where pt = "    "

    printParams :: [E] -> String
    printParams [] = ""
    printParams [x] = show x
    printParams (x:xs) = (show x)++", "++(printParams xs)
    
    isComp :: Statements -> Bool
    isComp (Composition _ _) = True
    isComp _ = False

    prettyPrintStatements :: Statements -> Int -> String
    prettyPrintStatements Null t =  (printTabs t)++"Null"
    prettyPrintStatements (Composition s1 s2) t = let tabs = (printTabs t) in
                          (prettyPrintStatements s1 t)++";\n"++(prettyPrintStatements s2 t)
    prettyPrintStatements (AssignVar v e) t = (printTabs t)++(show v)++" := "++(show e)
    prettyPrintStatements (AssignArr a lp e) t = (printTabs t)++(show a)++"("++(printParams lp)++") := "++(show e)
    prettyPrintStatements (IfThenElse b s1 s2) t = let 
                            tabs = printTabs t
                            t1 = t+1
                            tabs1 = printTabs t1
                            pst2 = (prettyPrintStatements s2 t1)++";\n"
                           in tabs++"if "++(show b)++ " then\n"++(prettyPrintStatements s1 t1)++";\n"++tabs++"else\n"++pst2++tabs++"end if"
    prettyPrintStatements (While as b s) t = let 
                        tabs = printTabs t
                        t1 = t+1
                        tabs1 = printTabs t1
                         in tabs++"while "++(show b)++"\n"++tabs++"--# "++(show as)++"\n"++tabs++"loop\n"++(prettyPrintStatements s t1)++";\n"++tabs++"end loop"
    prettyPrintStatements (If b s) t = let 
                        tabs = printTabs t
                        t1 = t+1
                       in
                        tabs++"if "++(show b)++" then\n"++(prettyPrintStatements s t1)++";\n"++tabs++"end if"
    prettyPrintStatements (IfElsIf b1 s1 b2 s2) t = let 
                            tabs = printTabs t
                            t1 = t+1
                            tabs1 = printTabs t1
                            t2 = t+2
                            tabs2 = printTabs t2
                       in
                        tabs++"if "++(show b1)++" then\n"++(prettyPrintStatements s1 t1)++";\n"++tabs1++"elsif "++(show b2)++" then\n"++(prettyPrintStatements s2 t2)++";\n"++tabs++"end if"
    prettyPrintStatements (IfElsIfElse b1 s1 b2 s2 sf) t = let 
                                tabs = printTabs t
                                t1 = t+1
                                tabs1 = printTabs t1
                                t2 = t+2
                                tabs2 = printTabs t2
                                t3 = t+3
                                tabs3 = printTabs t3
                              in
                                tabs++"if "++(show b1)++" then\n"++(prettyPrintStatements s1 t1)++";\n"++tabs1++"elsif "++(show b2)++" then\n"++(prettyPrintStatements s2 t2)++";\n"++tabs2++"else\n"++(prettyPrintStatements sf t3)++";\n"++tabs++"end if"
    prettyPrintStatements (InfiniteLoop s) t = let 
                        tabs = printTabs t
                        t1 = t+1
                          in tabs++"loop\n"++(prettyPrintStatements s t1)++";\n"++tabs++"end loop"
    prettyPrintStatements (LoopExitWhen1 s b) t = let 
                        tabs = printTabs t
                        t1 = t+1
                        tabs1 = printTabs t1
                          in tabs++"loop\n"++(prettyPrintStatements s t1)++";\n"++tabs1++"exit when "++(show b)++";\n"++tabs++"end loop"
    prettyPrintStatements (LoopExitWhen2 as b s) t = let 
                        tabs = printTabs t
                        t1 = t+1
                        tabs1 = printTabs t1
                         in tabs++"loop\n"++tabs++"--# "++(show as)++";\n"++"exit when "++(show b)++";\n"++(prettyPrintStatements s t1)++";\n"++tabs++"end loop"
    prettyPrintStatements (CallPF f) t = (printTabs t)++(show f)
    prettyPrintStatements (CallF f le) t = (printTabs t)++(show f)++"( "++ (printList le) ++" )"
    prettyPrintStatements (CallPP p) t = (printTabs t)++(show p)
    prettyPrintStatements (Execution_Block dt dv dsp s) t = let
                        tabs = printTabs t
                        t1 = t+1
                       in tabs++"types\n\n"++(prettyPrintStatements (Declaration_Type dt) t1)++"\n"
                        ++tabs++"variables\n\n"++(prettyPrintStatements (Declaration_Var dv) t1)++"\n"
                        ++tabs++"subprograms\n\n"++(prettyPrintStatements (Declaration_Subprogram dsp) t1)++"\n"
                        ++tabs++"execute\n\n"++(prettyPrintStatements s t1)++";\n\n"
                        ++"end"
    prettyPrintStatements (Declaration_Type dt) t = prettyPrintTypes dt t
    prettyPrintStatements (Declaration_Var dv) t = prettyPrintDV dv t --(show dv)
--    prettyPrintStatements (Declaration_Function df) t = prettyPrintFunctions df t
    prettyPrintStatements (Declaration_Subprogram (D_spf df)) t = prettyPrintFunctions df t
    prettyPrintStatements (Declaration_Subprogram (D_spp dp)) t = prettyPrintProcedures dp t
--		      IfListElsIf BE Statements [(BE, Statements)] |
--		      IfListElsIfElse BE Statements [(BE, Statements)] Statements |

    prettyPrintTypes :: D_t -> Int -> String
    prettyPrintTypes End_D_t _ = ""
    prettyPrintTypes (Dt ty r dt) t = (printTabs t)++"type "++(show ty)++" is "++(show r)++";\n"++(prettyPrintTypes dt t)
    prettyPrintTypes (Dst ty1 ty2 r dt) t = (printTabs t)++"subtype "++(show ty1)++" is "++(show ty2)++" "++(show r)++";\n"++(prettyPrintTypes dt t)


    prettyPrintDV :: D_v -> Int -> String
    prettyPrintDV End_D_v t = ""
    prettyPrintDV (Dv [] _ d) t = prettyPrintDV d t
    prettyPrintDV (Dv lv ty d) t = (printTabs t) ++ (printList lv) ++ " : "++(show ty)++";\n" ++ (prettyPrintDV d t)
    prettyPrintDV (DAv [] ty cae d) t = prettyPrintDV d t
    prettyPrintDV (DAv lv ty cae d) t = (printTabs t) ++ (printList lv) ++ " : "++(show ty)++" := "++(show cae)++";\n" ++ (prettyPrintDV d t)

    prettyPrintFunctions :: D_f -> Int -> String
    prettyPrintFunctions End_D_f t = ""
    prettyPrintFunctions (Dpf f c rt dv dfn s re dfe) t = (printTabs t)++"function "++(show f)++" return "++(show rt)++" is\n"
							++ (prettyPrintDV dv (t+1))++(prettyPrintFunctions dfn (t+1))++"\n"
							++(printTabs t)++"begin\n\n"++(prettyPrintStatements s (t+1))
							++ ";\n" ++ (printTabs (t+1)) ++ "return "++(show re)++";\n\n"++(printTabs t)++"end "++(show f)++";\n"++(prettyPrintFunctions dfe t)
    prettyPrintFunctions (Df f pf c rt dv dfn s re dfe) t  = (printTabs t)++"function "++(show f)++"( "++params++" )"++" return "++(show rt)++" is\n"
							    ++ (prettyPrintDV dv (t+1))++(prettyPrintFunctions dfn (t+1))++"\n"
							    ++(printTabs t)++"begin\n\n"++(prettyPrintStatements s (t+1))
							    ++ ";\n" ++ (printTabs (t+1)) ++ "return "++(show re)++";\n\n"++(printTabs t)++"end "++(show f)++";\n"++(prettyPrintFunctions dfe t)
							 where params = printListParameters pf
							       printListParameters ([x], V vf) = let (Just (ty, _)) = vf x in
												  (show x)++" : "++(show ty)
							       printListParameters (x:xs, V vf) = let (Just (ty, _)) = vf x in
												  (show x)++" : "++(show ty)++"; "++(printListParameters (xs, V vf))

    prettyPrintProcedures :: D_p -> Int -> String
    prettyPrintProcedures End_D_p t = ""
    prettyPrintProcedures (Dpp p c dv dspn s dspe) t = (printTabs t)++"procedure "++(show p)++" is\n"
						     ++ (prettyPrintDV dv (t+1))++(prettyPrintStatements (Declaration_Subprogram dspn) (t+1))++"\n"
						     ++ (printTabs t)++"begin\n\n"++(prettyPrintStatements s (t+1))
							++ ";\n\n"++(printTabs t)++"end "++(show p)++";\n"++(prettyPrintStatements (Declaration_Subprogram dspe) t)

    printList :: (Show x) => List x -> String
    printList [x] = show x
    printList (x:xs) = (show x)++", "++(printList xs)


    data D_v = End_D_v |
           Dv [ID_v] ID_t D_v |
           DAv [ID_v] ID_t CAE D_v
        deriving Show

    data D_f = End_D_f |
           Dpf ID_f ContractSpec ID_t D_v D_f Statements E D_f |
           Df ID_f Parameters_f ContractSpec ID_t D_v D_f Statements E D_f
        deriving Show

    data D_p = End_D_p |
           Dpp ID_p ContractSpec D_v D_sp Statements D_sp
        deriving Show

    data D_sp = D_spf D_f | D_spp D_p
        deriving Show

    data D_t = End_D_t | Dt ID_t R D_t | Dst ID_t ID_t R D_t
        deriving Show

    data D_r = D_r
        deriving Show

